@@ -7,13 +7,13 @@
 __package_info__ = """
 BEGIN PYTHON-PACKAGE-INFO 1.0
 Title:			mxStack - a fast stack implementation
-Current-Version:	0.2
+Current-Version:	0.3.0
 Home-Page:		http://starship.skyport.net/~lemburg/mxStack.html
-Primary-Site:		http://starship.skyport.net/~lemburg/mxStack-0.2.zip
+Primary-Site:		http://starship.skyport.net/~lemburg/mxStack-0.3.0.zip
 
 This package contains a generic and very simple stack implementation
 written in C for speed. For full documentation see the home page.
 END PYTHON-PACKAGE-INFO
 """
-from mxStack.mxStack import *
-from mxStack.mxStack import __version__
+from mxStack import *
+from mxStack import __version__
