@@ -9,13 +9,18 @@
 /*#define MAL_MEM_DEBUG*/
 /*#define MAL_DEBUG*/
 
+/* Logging file used by debugging facility */
+#ifndef MAL_DEBUG_OUTPUTFILE
+# define MAL_DEBUG_OUTPUTFILE "mxStack.log"
+#endif
+
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXSTACK
 
 #include "mx.h"
 #include "mxStack.h"
 
-#define VERSION "0.2"
+#define VERSION "0.3.0"
 
 /* The minimal size of stacks; must be > 1 */
 #define MINIMAL_STACKSIZE	4
@@ -35,7 +40,7 @@ static char *Module_docstring =
 
  MXSTACK_MODULE" -- A stack implementation. Version "VERSION"\n\n"
 
- "(c) Copyright Marc-Andre Lemburg, 1997, mal@lemburg.com,\n\n"
+ "(c) Copyright Marc-Andre Lemburg, 1997, 1998, mal@lemburg.com,\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -98,6 +103,7 @@ PyObject *notimplemented1(PyObject *v)
 {
     Py_Error(PyExc_TypeError,
 	     "operation not implemented");
+
  onError:
     return NULL;
 }
@@ -152,7 +158,7 @@ mxStackObject *mxStack_New(int size)
 /* --- deallocation --- */
 
 static
-void mxStack_Free(mxStackObject *stack)
+void mxStack_Free(register mxStackObject *stack)
 {
     if (stack->array) {
 	int i;
@@ -169,63 +175,10 @@ void mxStack_Free(mxStackObject *stack)
 /* --- API functions --- */
 
 static
-mxStackObject *mxStack_FromSequence(PyObject *v)
-{
-    mxStackObject *stack = 0;
-    PyObject **w;
-    int size,length;
-    register int i;
-    
-    Py_Assert(PySequence_Check(v),
-	      PyExc_TypeError,
-	      "argument must be a sequence");
-    length = PySequence_Length(v);
-    Py_Assert(length >= 0,
-	      PyExc_TypeError,
-	      "sequence must have a __len__ method");
-
-    stack = PyObject_NEW(mxStackObject,&mxStack_Type);
-    if (stack == NULL)
-	goto onError;
-
-    /* Init vars */
-    size = length;
-    GROW(size);
-    if (size < MINIMAL_STACKSIZE)
-	size = MINIMAL_STACKSIZE;
-    w = new(PyObject*,size);
-    if (w == NULL) {
-	PyErr_NoMemory();
-	return NULL;
-    }
-    stack->array = w;
-    stack->size = size;
-    stack->top = -1;
-
-    /* Insert items */
-    for (i = 0; i < length; i++, stack->top++) {
-	register PyObject *o;
-	
-	o = PySequence_GetItem(v,i);
-	if (o == NULL)
-	    goto onError;
-	Py_INCREF(o);
-	w[i] = o;
-    }
-
-    return stack;
-
- onError:
-    if (stack)
-	mxStack_Free(stack);
-    return NULL;
-}
-
-static
-int mxStack_Push(mxStackObject *stack,
+int mxStack_Push(register mxStackObject *stack,
 		 PyObject *v)
 {
-    register int top;
+    int top;
     
     if (stack == NULL) {
 	PyErr_BadInternalCall();
@@ -259,7 +212,76 @@ int mxStack_Push(mxStackObject *stack,
 }
 
 static
-PyObject *mxStack_Pop(mxStackObject *stack)
+int mxStack_PushMany(register mxStackObject *stack,
+		     PyObject *seq)
+{
+    int top;
+    register int i;
+    int length;
+
+    if (stack == NULL) {
+	PyErr_BadInternalCall();
+	goto onError;
+    }
+
+    length = PySequence_Length(seq);
+    if (length < 0)
+	goto onError;
+
+    top = stack->top;
+
+    /* Grow the stack, if we hit the array boundary */
+    if (top + length >= stack->size) {
+	register PyObject **w;
+	register int size;
+
+	size = stack->size;
+	while (top + length >= size)
+	    GROW(size);
+	w = resize(stack->array,PyObject*,size);
+	if (w == NULL) {
+	    PyErr_NoMemory();
+	    goto onError;
+	}
+	stack->array = w;
+	stack->size = size;
+    }
+
+    /* Push the entries from left to right onto the stack */
+    for (i = 0; i < length; i++) {
+	register PyObject *v;
+	
+	if (PyTuple_Check(seq)) {
+	    v = PyTuple_GET_ITEM(seq,i);
+	    Py_INCREF(v);
+	}
+	else if (PyList_Check(seq)) {
+	    v = PyList_GET_ITEM(seq,i);
+	    Py_INCREF(v);
+	}
+	else {
+	    v = PySequence_GetItem(seq,i);
+	    if (!v) {
+		/* Rollback */
+		while (i > 0) {
+		    Py_DECREF(stack->array[top--]);
+		    i--;
+		}
+		stack->top = top;
+		goto onError;
+	    }
+	}
+	stack->array[++top] = v;
+    }
+    stack->top = top;
+    return 0;
+
+ onError:
+    return -1;
+}
+
+static
+PyObject *mxStack_Pop(register mxStackObject *stack)
 {
     if (stack == NULL) {
 	PyErr_BadInternalCall();
@@ -273,11 +295,11 @@ PyObject *mxStack_Pop(mxStackObject *stack)
     return NULL;
 }
 
-/* Pop n entries into a Python tuple from the top of the stack; order
-   is top to bottom. */
+/* Pop up to n entries into a Python tuple from the top of the stack;
+   order is top to bottom. */
 
 static
-PyObject *mxStack_PopMany(mxStackObject *stack,
+PyObject *mxStack_PopMany(register mxStackObject *stack,
 			  int n)
 {
     PyObject *t;
@@ -287,9 +309,7 @@ PyObject *mxStack_PopMany(mxStackObject *stack,
 	PyErr_BadInternalCall();
 	goto onError;
     }
-    Py_Assert(stack->top >= n - 1,
-	      mxStack_Error,
-	      "stack doesn't have enough entries");
+    n = min(stack->top + 1, n);
     t = PyTuple_New(n);
     if (!t)
 	goto onError;
@@ -301,7 +321,40 @@ PyObject *mxStack_PopMany(mxStackObject *stack,
 }
 
 static
-PyObject *mxStack_AsTuple(mxStackObject *stack)
+int mxStack_Clear(register mxStackObject *stack)
+{
+    register int i;
+	
+    for (i = 0; i <= stack->top; i++)
+	Py_DECREF(stack->array[i]);
+    stack->top = -1;
+    return 0;
+}
+
+static
+mxStackObject *mxStack_FromSequence(PyObject *seq)
+{
+    mxStackObject *stack = 0;
+    
+    /* Create an "empty" stack */
+    stack = mxStack_New(0);
+    if (stack == NULL)
+	goto onError;
+
+    /* Insert items */
+    if (mxStack_PushMany(stack,seq))
+	goto onError;
+
+    return stack;
+
+ onError:
+    if (stack)
+	mxStack_Free(stack);
+    return NULL;
+}
+
+static
+PyObject *mxStack_AsTuple(register mxStackObject *stack)
 {
     PyObject *t = 0;
     int i,len;
@@ -323,13 +376,14 @@ PyObject *mxStack_AsTuple(mxStackObject *stack)
     }
     return t;
  onError:
-    if (t)
+    if (t) {
 	Py_DECREF(t);
+    }
     return NULL;
 }
 
 static
-PyObject *mxStack_AsList(mxStackObject *stack)
+PyObject *mxStack_AsList(register mxStackObject *stack)
 {
     PyObject *l = 0;
     int i,len;
@@ -351,11 +405,48 @@ PyObject *mxStack_AsList(mxStackObject *stack)
     }
     return l;
  onError:
-    if (l)
+    if (l) {
 	Py_DECREF(l);
+    }
     return NULL;
 }
 
+static
+int mxStack_Length(register mxStackObject *self)
+{
+    return self->top + 1;
+}
+
+static
+int mxStack_Resize(register mxStackObject *self,
+		   int size)
+{
+    register PyObject **w;
+
+    if (self == NULL) {
+	PyErr_BadInternalCall();
+	goto onError;
+    }
+    
+    if (size < self->top)
+	size = self->top + 1;
+    if (size < MINIMAL_STACKSIZE)
+	size = MINIMAL_STACKSIZE;
+    
+    GROW(size);
+    w = resize(self->array,PyObject*,size);
+    if (w == NULL) {
+	PyErr_NoMemory();
+	goto onError;
+    }
+    self->array = w;
+    self->size = size;
+
+    return 0;
+
+ onError:
+    return -1;
+}
 
 /* --- methods --- (should have lowercase extension) */
 
@@ -373,6 +464,7 @@ Py_C_Function( mxStack_new,
     if (v == NULL)
 	goto onError;
     return v;
+
  onError:
     return NULL;
 }
@@ -387,6 +479,22 @@ Py_C_Function( mxStack_push,
     if (mxStack_Push(stack,v))
 	goto onError;
     Py_ReturnNone();
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxStack_push_many,
+	       "push_many(sequence)")
+{
+    PyObject *seq;
+
+    Py_GetSingleArg("O",seq);
+    
+    if (mxStack_PushMany(stack,seq))
+	goto onError;
+    Py_ReturnNone();
+
  onError:
     return NULL;
 }
@@ -402,6 +510,7 @@ Py_C_Function( mxStack_pop,
     if (v == NULL)
 	goto onError;
     return v;
+
  onError:
     return NULL;
 }
@@ -418,6 +527,20 @@ Py_C_Function( mxStack_pop_many,
     if (v == NULL)
 	goto onError;
     return v;
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxStack_clear,
+	       "clear()")
+{
+    Py_NoArgsCheck();
+    
+    if (mxStack_Clear(stack))
+	goto onError;
+    Py_ReturnNone();
+
  onError:
     return NULL;
 }
@@ -433,6 +556,7 @@ Py_C_Function( mxStack_as_tuple,
     if (v == NULL)
 	goto onError;
     return v;
+
  onError:
     return NULL;
 }
@@ -448,6 +572,23 @@ Py_C_Function( mxStack_as_list,
     if (v == NULL)
 	goto onError;
     return v;
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxStack_resize,
+	       "resize([size=len(stack)])")
+{
+    int size = mxStack_Length(stack);
+
+    Py_GetArg("|i",size);
+    
+    if (mxStack_Resize(stack,size))
+	goto onError;
+
+    Py_ReturnNone();
+
  onError:
     return NULL;
 }
@@ -509,16 +650,10 @@ PyObject *mxStack_Getattr(mxStackObject *self,
 			 (PyObject *)self,name);
 }
 
-static
-int mxStack_Length(mxStackObject *self)
-{
-    return self->top + 1;
-}
-
 static
 int mxStack_NonZero(mxStackObject *self)
 {
-    return self->top > 0;
+    return self->top >= 0;
 }
 
 /* Undocumented feature:
@@ -671,9 +806,12 @@ PyMethodDef mxStack_Methods[] =
 {   
     Py_MethodListEntrySingleArg("push",mxStack_push),
     Py_MethodListEntryNoArgs("pop",mxStack_pop),
+    Py_MethodListEntrySingleArg("push_many",mxStack_push_many),
     Py_MethodListEntrySingleArg("pop_many",mxStack_pop_many),
     Py_MethodListEntryNoArgs("as_tuple",mxStack_as_tuple),
     Py_MethodListEntryNoArgs("as_list",mxStack_as_list),
+    Py_MethodListEntryNoArgs("clear",mxStack_clear),
+    Py_MethodListEntry("resize",mxStack_resize),
     {NULL,NULL} /* end of list */
 };
 
@@ -720,7 +858,11 @@ mxStackModule_APIObject mxStackModuleAPI =
     mxStack_Push,
     mxStack_Pop,
     mxStack_AsTuple,
-    mxStack_AsList
+    mxStack_AsList,
+    mxStack_PopMany,
+    mxStack_PushMany,
+    mxStack_Clear,
+    mxStack_Length
 };
 
 /* create PyMethodObjects and register them in the module's dict */
@@ -761,10 +903,9 @@ MX_EXPORT(void)
     PyDict_SetItemString(moddict,MXSTACK_MODULE"API",api);
     Py_DECREF(api);
 
-    /* Check for errors */
+ onError:
+    /* Check for errors and report them */
     if (PyErr_Occurred())
-	Py_Error(PyExc_ImportError,
-		 "initialization of module "MXSTACK_MODULE" failed");
-onError:
+	Py_ReportModuleInitError(MXSTACK_MODULE);
     return;
 }
